#include "FrequentStrings.h"

void FrequentStrings::print(CSA* csa, ulong p, ulong l) {
//     printf("oikea SA[%d] = %lu\n", p, csa->lookup(p));
    uchar *result = csa->substring(csa->lookup(p), l);
    for (unsigned i = 0; (i < l) && ((char) result[i] > 1); i++)
        cout << (char) result[i];
    cout << endl;
    delete [] result;
}

// This method assumes the LCP-array is LCP[0..length] and
// LCP[length] = LCP[0] = -1. The SA is SA[0..length-1].
void FrequentStrings::frequent(char* filename1, char* filename2, int format, char query, double min, double max) 
{
//     if (min < 1) {
//         cout << "Minimum frequency must be >= 1!" << endl;
//         return;
//     }
    std::cerr << "Min " << min << ", Max " << max << std::endl;
//     TestDataStructures(); exit(0);


    /**
     * FIRST PHASE: PREPROCESSING
    **/
    Tools::StartTimer();
    CSAfor2DB* SA4DB = new CSAfor2DB(filename1, filename2, format);
    ulong length = SA4DB->length + 1; // length of LCP (not counting last -1)
    cerr << "anz1 = " << SA4DB->anz1 << ", anz2 = " << SA4DB->anz2 << ", anz = " << SA4DB->anz <<  endl;

    if (min*SA4DB->anz1 < 1) {
        cerr << "Support must be at least 1/|db_1|... exit!" << endl;
        exit(0);
    }
    if (max*SA4DB->anz2 < 1) {
        cerr << "Support must be at least 1/|db_2|... exit!" << endl;
        exit(0);
    }
//     for (unsigned ka = 0; ka < length; ka++)
//          printf("SA[%u] = %lu\n", ka, SA4DB->csa->lookup(ka));

    CHgtArray* hgtA = new CHgtArray(SA4DB->csa, SA4DB->t, length);
    SA4DB->DeleteText();
    
//     for (ulong i = length - 10; i <= length; i ++)
//         cerr << "LCP[" << i << "] = " << hgtA->GetPos(i) << endl;
    
    /**
     * SECOND PHASE: LABELLING
     **/
    unsigned labelWidth = Tools::CeilLog2(SA4DB->anz+1);
    ulong *label = new ulong[labelWidth * length / W + 1];
    Tools::SetField(label, labelWidth, 0, 0);
    ulong k = SA4DB->anz;
    ulong j = SA4DB->csa->GetBwtEndPosition();
    j = SA4DB->csa->LFmapping(j);
    for (ulong i = length - 1; i > 0; i--) {
        uchar c = SA4DB->csa->charAtPos(i);
        if (c == '\1')
            k --;
        Tools::SetField(label, labelWidth, j, k);

        j = SA4DB->csa->LFmapping(j);
    }
    Tools::SetField(label, labelWidth, j, k);
    if (k != 0) // Assert
    {
        cerr << "Labeling failed, k = " << k << endl;
        exit(0);
    }

    WaveletTree *wt = new WaveletTree(label, length, labelWidth);
    
    // Sanity check for WT
    /*int *prev = new int[SA4DB->anz+1];
    ulong *count = new ulong[SA4DB->anz+1];
    for (ulong i= 0; i <= SA4DB->anz; i ++) {
        prev[i] = -1;
        count[i] = 0;
    }
    
    for (ulong i = 0; i < length; i ++)
    {
        ulong ee = Tools::GetField(label, labelWidth, i);
        
//         cerr << "i = " << i << ", ee = " << ee << ", rank = " << wt->rank(ee,i) << endl;
        if (prev[ee] == -1)
        {
            if (wt->rank(ee,i) != 1)
            {
                cerr << "i = " << i << ", ee = " << ee << ", rank = " << wt->rank(ee,i) << endl;
                cerr << "WT->RANK == 1 test failed\n";
                exit(0);
            }
        }
        else
        {
            ulong pp = wt->rank(ee,i) - 1;
            if (pp != count[ee])
            {
                cerr << "count = " << count[ee] << " and pp = " << pp << "\n";
                exit(0);
            }
            if ((unsigned)prev[ee] != wt->select(ee, pp))
            {
                cerr << "prev = " << prev[ee] << " and wt = " << wt->select(ee, pp) << "\n";
                exit(0);
            }
        }
        prev[ee] = i;
        count[ee] ++;
    }
    delete [] prev;
    delete [] count;
    cerr << "WT check done.\n";*/

    delete [] label;
    label = 0;
    
    RMQ_succinct* rmq = new RMQ_succinct(hgtA, length);

//     for (unsigned ka = 0; ka < length; ka++)
//          printf("LCP[%u] = %lu\n", ka, hgtA->GetPos(ka));


    ulong entry; // # of entry in DB
    ulong pos;   // position of RMQ

    std::cerr << "Construction " << Tools::GetTime() << " s." << std::endl << "Extracting..." << std::endl;
    /**
     * THIRD PHASE: EXTRACTION
     **/
    ulong olds1, olds2;           // s-values of last completely visited subtree
    ulong oldc1, oldc2;           // c-values of last completely visited subtree
    ulong h_parent;               // stores height of parent node (for output)
    double supp1, supp2, growthrate;

    // Init
    R = new StringDepth(length);        // Node string depths
    S1 = new DeltaArray(length);  // number of leaves from DB1 in subtree
    S2 = new DeltaArray(length);  // number of leaves from DB2 in subtree
    C1 = new CorrectionFactor();  // correction factor for DB1
    C2 = new CorrectionFactor();  // correction factor for DB2

    // Init root node
    S1->Add(0);
    S2->Add(0);

    // start at 1 to skip the end-marker!
    for (ulong k = 1; k <= length; k++) 
    { 
        olds1 = 0; olds2 = 0;
        oldc1 = 0; oldc2 = 0;

        // String depth of top node
        ulong node_h = R->Top();
        // LCP
        ulong lcp_k = hgtA->GetPos(k);

        while (node_h > lcp_k) {
            olds1 = S1->Top();
            olds2 = S2->Top();
            oldc1 = C1->Top();
            oldc2 = C2->Top();

            supp1 = ((double)(olds1 - oldc1)) / ((double)SA4DB->anz1);
            supp2 = ((double)(olds2 - oldc2)) / ((double)SA4DB->anz2);
            if (supp2 != 0)
                growthrate = supp1 / supp2;
            else
                growthrate = max;


//             printf("k = %lu, s1 = %lu, s2 = %lu, c1 = %lu, c2 = %lu, node_h = %lu, node_l = %lu\n", k, olds1, olds2, oldc1, oldc2, node_h, k-1);

            // String depth of next node is:
            h_parent = R->Parent();
            if (h_parent >= lcp_k)
            {
                // propagate counters upwards:
                ulong sumS = S1->Sum(2);
                S1->Remove(2);
                S1->Add(sumS);
                sumS = S2->Sum(2);
                S2->Remove(2);
                S2->Add(sumS);

                C1->SumTopNodes();
                C2->SumTopNodes();
            }
            else 
                h_parent = lcp_k;

            // Check for frequent strings  OR  emerging substrings
            if ((query == 'F' && supp1 >= min && supp2 <= max)
                || (query == 'E' && supp1 >= min && growthrate >= max))
            {
                //SA4DB->t[(int)SA4DB->csa->lookup(left)+h_parent] > '\1' &&
                //SA4DB->t[SA4DB->csa->lookup(left)+hgtA->GetPos(k)] > '\1') { // This string is frequent!
                    // calculate true length of string s to be printed
                    // (time for loop is O(|s|))
                for (ulong h = 0; h < node_h; h++) {
                    if (SA4DB->csa->charAtPos(SA4DB->csa->lookup(k-1)+h) <= '\1') {
                        node_h = h; break;
                    }
                }
//                 printf("Printing: node_h = %lu, node_l = %lu, h_parent = %lu\n", node_h, left, h_parent);
                for (ulong h = h_parent + 1; h <= node_h; h++) {
                    print(SA4DB->csa, k-1, h);
                }
            }
            
            
            
//             printf("left = %lu, (%lu, %lu, %lu, %lu)\n", left,olds1,olds2,oldc1,oldc2);

            // Remove top node
            R->Pop();
            // String depth of top node:
            node_h = R->Top();
        }


//         printf("node->h = %d, LCP = %d\n", node_h, hgtA->GetPos(k));
        
        // Push LCA (every inner node IS lca of consec. leaves!)
        if (k == 1)
            PushNewNode(lcp_k, olds1, olds2, oldc1, oldc2);
        if (node_h < lcp_k) 
            R->Push(lcp_k);
            //PushNewNode(lcp_k, olds1, olds2, oldc1, oldc2);

        // Accumulate correction factors
        if (k < length) 
        {
            entry = wt->charAtPos(k); // FIXME
            pos = wt->rank(entry, k) - 1;
            if (pos > 0)
            {
                pos = wt->select(entry, pos); // Last entry position
                pos = rmq->query(pos+1, k);
                pos = hgtA->GetPos(pos);

                if (entry < SA4DB->anz1) 
                {
                    pos = R->FindDepth(pos);
                    C1->AddByOne(pos);
                }
                else
                {
                    pos = R->FindDepth(pos);
                    C2->AddByOne(pos);
                }
            }

            // push leaf:
            pos = entry < SA4DB->anz1 ? 1 : 0;
            PushNewNode(length - SA4DB->csa->lookup(k), pos, 1-pos, 0, 0);
        }
    }

    delete wt;
    delete hgtA;  // Disabled for memusage monitoring
    delete rmq;
    delete SA4DB;
    delete R;
    delete S1;
    delete S2;
    delete C1;
    delete C2;
}

void FrequentStrings::PushNewNode(ulong depth, ulong olds1, ulong olds2, ulong oldc1, ulong oldc2)
{
//      printf("Adding node (%lu, %lu, %lu, %lu, %lu)\n", depth, olds1, olds2, oldc1, oldc2);
    S1->Add(olds1);
    S2->Add(olds2);
    C1->Push(oldc1);
    C2->Push(oldc2);
    R->Push(depth);
    
//     printf("Added node  (%lu, %lu, %lu, %lu, %lu)\n", R->Top(), S1->Top(), S2->Top(), (C1->Top()), (C2->Top()));
}

void FrequentStrings::TestDataStructures()
{
    cerr << "Testing ..." << endl;
    // Init
    R = new StringDepth(10);        // Node string depths
    R->Push(0);
    R->Push(10);
    cerr << "Depth " << R->Top() << ", Parent " << R->Parent() << endl;
    R->Pop();
    cerr << "Depth " << R->Top() << ", Parent " << R->Parent() << endl;
    R->Push(5);
    R->Push(6);
    cerr << "Depth " << R->Top() << ", Parent " << R->Parent() << endl;
    cerr << "Found " << R->FindDepth(0) << endl; 
    R->PrintVector();
    
    
    
    S1 = new DeltaArray(10);  // number of leaves from DB1 in subtree
    S2 = new DeltaArray(10);  // number of leaves from DB2 in subtree
    C1 = new CorrectionFactor();  // correction factor for DB1
    C1->Push(0);
    C1->Push(10);
    cerr << "Corr " << C1->Top() << endl;
    C1->Pop();
    cerr << "Corr " << C1->Top()  << endl;
    C1->Push(5);
    C1->Push(6);
    cerr << "Corr " << C1->Top() << endl;
    C1->AddByOne(2);C1->AddByOne(2);
    C1->SumTopNodes();
    cerr << "Corr " << C1->Top() << endl;
    C1->PrintVector();
    C1->SumTopNodes();
    C1->PrintVector();
    C1->SumTopNodes();
    C1->PrintVector();
    
    
    cerr << "Done." << endl;
}

////////////////////////////////////////////////////////////////////////////
// Class FrequentStrings::DeltaArray
FrequentStrings::DeltaArray::DeltaArray(ulong size)
{    
    // pathological cases size should be 7*size 
    // this is taken into account when startPos gets too small:
    // the memory size is doubled
    this->size = size / ((ulong)log2(size)*(ulong)log2(size));
    if (this->size < 1024) this->size = 1024;
    A = new ulong[this->size / W + 1];
    startPos = this->size - 1;
}

FrequentStrings::DeltaArray::~DeltaArray()
{
    delete [] A;
}

void FrequentStrings::DeltaArray::Add(ulong value)
{
    value ++;   // In this case, we want to code 0 also
    ulong l = Tools::bits(value);
    ulong m = Tools::bits(l);
    if (startPos < 2*m+1+l) {
       // we will run out of space, let's double the space
       // this is not tested yet completely...
       std::cerr << "-------------------doubling space for deltas\n";
       ulong *B;
       B = A;
       A = new ulong[size/W + 1 + size/W +1];
       for (ulong i=0; i <= size/W; i++)
          A[size/W+i+1]=B[i];
       delete [] B;
       startPos = (size/W+1)*W+startPos;
       size = (size/W+1)*W+size;    
    }
    startPos -= l;
    Tools::SetVariableField(A, l, startPos, value);
    startPos -= m;
    Tools::SetVariableField(A, m, startPos, l);
    startPos -= m + 1;
    Tools::SetVariableField(A, m + 1, startPos, (1 << m) - 1);
}
    
void FrequentStrings::DeltaArray::Remove(ulong n)
{
    for (ulong i = 0; i < n; i++)
    {
        ulong k = 0;
        while (Tools::GetVariableField(A, 1, startPos + k))
            k++;
        startPos += k + 1;
        ulong j = Tools::GetVariableField(A, k, startPos);
        startPos += k + j;
    }
}

// Returns sum of n values, and the index after the last value (see GetNext())
ulong FrequentStrings::DeltaArray::Sum(ulong n)
{
    ulong sum = 0;
    ulong index = startPos; // Reset index parameter
    
    for (ulong i = 0; i < n; i++)
    {
        ulong k = 0;
        while (Tools::GetVariableField(A,1,index+k)) 
            k++;
        index += k + 1;   
        ulong j = Tools::GetVariableField(A, k, index);
        index += k;
        sum += Tools::GetVariableField(A, j, index) - 1;
        index += j;
    }
    return sum;
}

// Returns the first value
ulong FrequentStrings::DeltaArray::Top()
{
    ulong index = startPos; // Reset index parameter
    ulong k = 0;
    while (Tools::GetVariableField(A,1,index+k)) 
        k++;
    index += k + 1;   
    ulong j = Tools::GetVariableField(A, k, index);
    index += k;
    ulong sum = Tools::GetVariableField(A, j, index) - 1;
    index += j;

    return sum;
}

////////////////////////////////////////////////////////////////////////////
// Class FrequentStrings::StringDepth
FrequentStrings::StringDepth::StringDepth(ulong length)
{
    R = new BVTree();                       // String depths on right-most path
    R->appendBit(true);                     // String depth of the root node == 0
    for (ulong i = 0; i < length+1; i++)
        R->appendBit(false);                // Fill with n zero-bits
    nodeCount = 1;                          // number of nodes in the "stack"
    topIndex = 1;                           // Index of the top node
    prevIndex = 0;                          // Index of the parent of the top node
}

FrequentStrings::StringDepth::~StringDepth()
{
    delete R;
    R = 0;
}

void FrequentStrings::StringDepth::PrintVector()
{
    for (ulong i = 1; i <= topIndex; i ++)
        if ((*R)[i])
            cerr << "1";
        else
            cerr << "0";
    cerr << endl;
    for (ulong i = 1; i <= topIndex; i ++) {
        if (i == topIndex)
            cerr << "^";
        else if (i == prevIndex)
            cerr << "^";
        else cerr << " ";
    }
    cerr << endl;
    cerr << "Top = " << topIndex << ", prev = " << prevIndex << endl;
}

ulong FrequentStrings::StringDepth::Top()
{
    return (topIndex - nodeCount);
}

// String depth of the parent of the top node
ulong FrequentStrings::StringDepth::Parent()
{
    return (prevIndex - nodeCount + 1);
}

void FrequentStrings::StringDepth::Push(ulong depth)
{
    ulong index = R->select0(depth+1);
    R->insertBit(true, index);
    nodeCount ++;
    prevIndex = topIndex;
    topIndex = index;
}

void FrequentStrings::StringDepth::Pop()
{
    R->deleteBit(topIndex);
    nodeCount --;
    topIndex = prevIndex;
    if (nodeCount > 1)
        prevIndex = R->select1(nodeCount - 1);
    else
        prevIndex = 0;
}

ulong FrequentStrings::StringDepth::FindDepth(ulong depth)
{
    depth = R->select0(depth+1);
    return R->rank1(depth);
}

////////////////////////////////////////////////////////////////////////////
// Class FrequentStrings::CorrectionFactor
FrequentStrings::CorrectionFactor::CorrectionFactor()
{
    C = new BVTree();
    C->appendBit(true);
    nodeCount = 1;                          // number of nodes in the "stack"
    topIndex = 1;                           // Index of the top node
    prevIndex = 0;                          // Index of the parent of the top node
}

FrequentStrings::CorrectionFactor::~CorrectionFactor()
{
    delete C;
    C = 0;
}

void FrequentStrings::CorrectionFactor::PrintVector()
{
    for (ulong i = 1; i <= topIndex; i ++)
        if ((*C)[i])
            cerr << "1";
    else
        cerr << "0";
    cerr << endl;
    for (ulong i = 1; i <= topIndex; i ++) {
        if (i == topIndex)
            cerr << "^";
        else if (i == prevIndex)
            cerr << "^";
        else cerr << " ";
    }
    cerr << endl;
    cerr << "Top = " << topIndex << ", prev = " << prevIndex << endl;
}

ulong FrequentStrings::CorrectionFactor::Top()
{
    return (topIndex - prevIndex - 1); // select1(nodeCount) - C1->select1(nodeCount-1) - 1;
}

// String depth of the parent of the top node
void FrequentStrings::CorrectionFactor::SumTopNodes()
{
//     printf("Sumtopnodes()\n");
    if (nodeCount <= 2)
    {
        cerr << "SumTopNodes() called with nodeCount = " << nodeCount << endl;
        exit(0);
    }
    topIndex --;
    C->deleteBit(prevIndex);
    nodeCount --;
    if (nodeCount > 1)
        prevIndex = C->select1(nodeCount-1);
    else
        prevIndex = 0;
}

void FrequentStrings::CorrectionFactor::Push(ulong depth)
{
//     printf("Push()\n");
    prevIndex = topIndex;
    topIndex += depth + 1;
    while (depth > 0)
    {
        depth --;
        C->appendBit(false);
    }
    C->appendBit(true);
    nodeCount ++;
}

void FrequentStrings::CorrectionFactor::AddByOne(ulong node)
{
//     printf("Addbyone()\n");
    node = C->select1(node);
    if (node <= prevIndex)
        prevIndex ++;

    topIndex ++;
    C->insertBit(false, node);
}

void FrequentStrings::CorrectionFactor::Pop()
{
//         printf("Pop()\n");
    for (ulong i = topIndex; i > prevIndex; i --)
        C->deleteBit(i);
    nodeCount --;
    topIndex = prevIndex;
    if (nodeCount > 1)
        prevIndex = C->select1(nodeCount-1);
    else
        prevIndex = 0;
}
